library ac_dart.host_service.http_client;

import 'package:atlassian_connect_host/host_event_model.dart';
import 'package:atlassian_connect_host/host_registration_persistence.dart';
import 'package:atlassian_connect_host/src/preconditions.dart';
import 'package:atlassian_connect_host/src/http_method.dart';
import 'dart:async';
import 'dart:io';
import 'package:path/path.dart' as _path;
import 'package:option/option.dart';
import 'package:logging/logging.dart';
import 'package:atlassian_connect_jwt/atlassian_connect_jwt.dart';
import 'package:quiver/collection.dart';



_path.Context get _url => _path.url;

final Logger _log = new Logger('AcHostHttpClient');

/**
 * An http client for communicating with Atlassian product hosts. Supports
 * signing requests with the required Jwt mechanism.
 * 
 * Note this is a wrapper for the core [HttpClient] class
 */
class AcHostHttpClient {
  final Option<String> _sharedSecret;
  final Option<String> _issuer;
  final Option<String> _audience;
  final Uri baseUri;
  final HttpClient _httpClient;

  AcHostHttpClient.fromHostRego(AcHostRegistration hostRego, AddonHost addonHost) 
    : this(checkNotNull(hostRego).hostDetails.baseUrl,
      new Some(hostRego.sharedSecret),
      new Some(addonHost.key),
      null,
      new Some(hostRego.hostDetails.key));
  
  /// [httpClient] exposed for testing
  AcHostHttpClient(this.baseUri, [ this._sharedSecret = const None(), 
      this._issuer = const None(),
      HttpClient httpClient,
      this._audience = const None()]) : _httpClient = httpClient != null ? 
          httpClient : new HttpClient();

  Future<HttpClientRequest> get(String path, {bool signRequest: false, 
    Map queryParams: const {}, String subject }) {
    return _httpClient.getUrl(
        _createUri(path, HttpMethod.GET, signRequest, queryParams, subject));
  }

  Future<HttpClientRequest> open(String path, HttpMethod method, 
      { bool signRequest: false, Map queryParams: const {}, String subject }) {
    return _httpClient.openUrl(method.literal, 
        _createUri(path, method, signRequest, queryParams, subject));
  }

  Future<HttpClientRequest> post(String path, {bool signRequest: false, 
    Map queryParams: const {}, String subject }) {
    return _httpClient.postUrl(
        _createUri(path, HttpMethod.POST, signRequest, queryParams, subject));
  }

  void addCredentials(String realm, HttpClientCredentials credentials) {
    _httpClient.addCredentials(baseUri, realm, credentials);
  }
  
  Uri _createUri(String path, HttpMethod method, bool signRequest, 
                 Map queryParams, String subject) {
    final Map params = signRequest ? new Map.from(queryParams) : queryParams;
    
    if (signRequest) {
      if (_sharedSecret.isEmpty() || _issuer.isEmpty()) {
        throw new ArgumentError("Cannot sign a request without a shared secret"
            " and issuer");
      }
      params['jwt'] = createProductHostToken(path, method.literal, queryParams, 
          _sharedSecret.get(), _issuer.get(), subject: subject,
          audience: _audience.getOrElse(() => null));
    }
    
    Uri uri = Uri.parse(_url.join(_url.fromUri(baseUri), path));
    if (params.isNotEmpty) {
      uri = uri.scheme == 'http' ? new Uri.http(uri.authority, uri.path, params) :
        new Uri.https(uri.authority, uri.path, params);
    }
    
    _log.finest('sending ${method.literal} $uri');
    return uri;
  }
  
  Multimap<String, String> _toMultiMap(Map m) {
    final mm = new Multimap();
    m.forEach((k, v) => mm.add(k, v));
    return mm;
  }

}

